#' Runs the SWAT+ soft-calibration routine for water balance components.
#'
#' \code{run_softcal_waterbalance} executes the SWAT+ soft calibration
#' routine to adjust the overall water balance of a SWAT+ model setup.
#' The user can specify the \code{wateryield_ratio} and the
#' \code{baseflow_ratio} of the simulated watershed. Based on the provided
#' ratios the routine adjusts the respective model parameters to improve
#' the fit of the simulated water balance components to the provided ratios.
#'
#' @param project_path Path to the SWAT+ project folder (i.e. TxtInOut).
#' @param wyr Water yield ratio; Numeric value between 0 and 1 which defines the
#'   fraction of precipitation which becomes water yield through fast, slow,
#'   and aquifer flow
#' @param bfr Base flow ratio; Numeric value between 0 and 1 which defines the
#'   fraction of water yield which is base flow
#' @param start_date (optional) Start date of the SWAT simulation. Provided as
#'   character string in any ymd format (e.g. 'yyyy-mm-dd'), numeric value
#'   in the form yyyymmdd, or in Date format.
#' @param end_date (optional) End date of the SWAT simulation. Provided as
#'   character string in any ymd format (e.g. 'yyyy-mm-dd'), numeric value
#'   in the form yyyymmdd, or in Date format.
#' @param years_skip (optional) Integer value to define the number of simulation
#'   years that are skipped before writing SWAT model outputs.
#' @param keep_folder (optional) If \code{keep_folder = TRUE}
#'   '.model_run/verification' is kept and not deleted after finishing model runs.
#'   In this case '.model_run' is reused in a new model run if \code{refresh = FALSE}.
#'   \code{Default = FALSE}
#'
#' @return Returns the simulation results for the soft calibration routine as a
#'   tibble.
#' @importFrom processx run
#' @importFrom tibble %>% tibble
#' @export
#'
run_softcal_waterbalance <- function(project_path, wyr, bfr,
                                     start_date = NULL, end_date = NULL,
                                     years_skip = NULL, keep_folder = FALSE) {
  # Check settings before starting to set up
  ## General function input checks
  stopifnot(is.character(project_path))
  stopifnot(is.numeric(wyr))
  stopifnot(is.numeric(bfr))
  stopifnot(is.logical(keep_folder))

  cat("Creating temp model directory...\n")
  # create a temporary directory copy of the model setup
  run_path <- build_model_run(project_path, '/.run_softcal_wb')

  set_print_prt(project_path, run_path, outputs = 'sft', years_skip)
  set_time_sim(project_path, run_path, start_date, end_date)

  # Build the water balance soft calibration input files and
  # write them into the path where the softcal will be run.
  build_wb_softcal_input(run_path, wyr, bfr)

  # enables the soft calibration routine
  toggle_sft(run_path, switch = "on", routine = "wb")

  # This function is replaced by build_wb_softcal_input()
  # modify_wb_parms(temp_directory)

  ## Get operating system and find SWAT executable file
  os <- get_os()
  swat_exe <- find_swat_exe(project_path = run_path, os = os)

  cat("Running SWAT+ with soft-calibration routine...\n")
  msg <- run(run_os(swat_exe, os),
             wd = run_path,
             error_on_status = FALSE)

  cat("Reading results...\n")
  # reads the results of the wb soft calibration
  basin_wb_aa <- read_wb_aa(run_path)

  softcal_plot <- plot_wb_softcal(basin_wb_aa, wyr, bfr)

  # read the new .hyd file so that it can be returned
  hydrology_cal_hyd <- read_tbl('hydrology-cal.hyd', run_path, n_skip = 2) %>%
    set_names(., tolower(names(.)))

  return_data <-list(results = basin_wb_aa %>% tibble(),
                     result_plot = softcal_plot,
                     parameters = hydrology_cal_hyd)

  # delete the temp directory if the user does not want to keep it.
  if (keep_folder == FALSE) {
    cat("Deleting temporary directory...\n")
    unlink(run_path, recursive = TRUE, force = TRUE)
  }

  cat("Returning results...\n")

  softcal_plot

  return(return_data)
}

#' Build/load the water balance soft calibration input files and write them
#' into the folder where the model is executed.
#'
#' @param run_path Path to the copy of SWAT+ project where the soft calibration
#'   will be executed
#' @param wyr Water yield ratio; Numeric value between 0 and 1 which defines the
#'   fraction of precipitation which becomes water yield through fast, slow,
#'   and aquifer flow
#' @param bfr Base flow ratio; Numeric value between 0 and 1 which defines the
#'   fraction of water yield which is base flow
#'
#' @importFrom readr read_lines write_lines
#'
#' @keywords internal
#'
build_wb_softcal_input <- function(run_path, wyr, bfr) {
  # Remove existing soft calibration files in run folder
  invisible(
   suppressWarnings(
     file.remove(paste0(run_path, c('/water_balance.sft',
                                    '/codes.sft',
                                    '/wb_parms.sft')))
   )
  )
  # Path where the SWATdoctR package was installed
  pkg_path <- system.file(package = "SWATdoctR")

  # Path to the water balance softcal input files
  wb_sft_path <- paste0(pkg_path, '/extdata/wb_softcal')

  # Build codes soft and write into the model run folder
  codes_sft <- c('codes.sft: written by SWATdoctR',
                 'hyd_hru  hyd_hrulte  plnt  sed  nut  ch_sed  ch_nut  res',
                 '      y           n     n    n    n       n       n    n')

  write_lines(codes_sft, paste0(run_path, '/codes.sft'))

  # Copy parameter input files into the model run folder
  # file.copy(paste0(wb_sft_path, '/codes.sft'), run_path)
  file.copy(paste0(wb_sft_path, '/wb_parms.sft'), run_path)

  # Build the header lines of the water balance softcal file
  #
  wb_sft <- c('water_balance.sft: written by SWATdoctR',
              '1',
              'name                   num',
              'basin                    1',
              'name  surq_rto  latq_rto  perc_rto    et_rto tileq_rto       pet       sed       wyr       bfr      solp')
  # wb_sft <- read_lines(paste0(wb_sft_path, '/water_balance.sft'), lazy = FALSE)

  # Generate value string which includes the water yield and baseflow ratios
  # wateryield_ratio = parameters[[which(names(parameters) == "wateryield_ratio")]]

  # baseflow_ratio = parameters[[which(names(parameters) == "baseflow_ratio")]]

  wb_sft_val <- paste(c('null', rep('         0', 7),
                        sprintf(rep('%10.3f', 2), c(wyr, bfr)),
                        '         0'),
                      collapse = '')

  # Build the water balance soft calibration input file
  wb_sft <- c(wb_sft, wb_sft_val)

  # Write the water balance soft calibration input into the model run folder
  write_lines(wb_sft, paste0(run_path, '/water_balance.sft'))
}


#' downloads the required .sft files from some source
#' printing just for diagnostics, maybe remove.
#'
#' @param path text string to (temporary) directory
#' @keywords internal
#' @importFrom dplyr %>%
#' @importFrom utils download.file
#'
download_sft_files <- function(path) {
  # "water_balance.sft" currently does not exist on bitbucket but needs to be downloaded!!
  # TEMP until the water_balance.sft file is hosted
  if ("water_balance.sft" %in% list.files(path) == FALSE) {
    stop("water_balance.sft needs to present in your directory TEMP: for now...")
  }

  # add a backslash onto the path, so its compatible with build_model_run()
  path = paste0(path, "/")

  # the desired files might be better passed as a parameter instead of hard coded here
  required_files = c("codes.sft", "wb_parms.sft")

  # The download repository for the sft files:
  # I hope this remains constant for all the files???
  # or can we host this on our own gitlab?
  # this could also be passed as a parameter
  source__online_directory = "https://bitbucket.org/blacklandgrasslandmodels/modular_swatplus/raw/3d2126fb115077b51be90526238c83cd79e3ef4c/database_files/"

  # find out which .sft files are missing and download them.
  for (file in required_files) {
    if (file %in% list.files(path) == FALSE) {
      paste(file, "not in your SWAT+ directory. downloading now...") %>% cat()

      # not sure how we want to handle error-prone code?
      try(download.file(
        url = paste0(source__online_directory, file),
        destfile = paste0(path, file)
      ))

      # check to see if the file was downloaded
      if (file %in% list.files(path)) {
        paste(file, "downloaded successfully") %>% cat()
      } else{
        paste(file, "NOT DOWNLOADED!") %>% cat()
      }
    } else{
      paste(file, "already exists in directory, not downloading again") %>% cat()
    }
  }
}

#' toggles the soft calibration routine either ON or OFF
#'
#' @param path string to (temporary) directory
#' @param switch string, either "off" or "on"
#' @param routine string, either "wb" or "cy"
#' @keywords internal
#' @importFrom dplyr %>%
#'
toggle_sft <- function(path, switch, routine) {
  # add a backslash onto the path, so its compatible with build_model_run()
  path = paste0(path, "/")

  # file.cio modification:
  # read the file.cio in
  file_cio = readLines(con = paste0(path, "file.cio"))

  # grab the 22nd line and split that line based of white space
  line22 = file_cio[22] %>% base::strsplit("\\s+") %>% unlist()

  if (switch == "on") {
    if (routine == "wb") {
      # replace the the column values 4,5,6 with the .sft file names
      line22[4] = "codes.sft"
      line22[5] = "wb_parms.sft"
      line22[6] = "water_balance.sft"
      line22[9] = "null"
      line22[10] = "null"
      cat("Water balance soft cal enabled...\n")
    }
    if (routine == "cy") {
      line22[4] = "codes.sft"
      line22[5] = "null"
      line22[6] = "null"
      line22[9] = "plant_parms.sft"
      line22[10] = "plant_gro.sft"
      cat("Crop yield soft cal enabled...\n")
    }

  }

  if (switch == "off") {
    # replace the the column values 4,5,6 with NULL to disable
    line22[4] = "null"
    line22[5] = "null"
    line22[6] = "null"
    line22[9] = "null"
    line22[10] = "null"
    cat("Soft cal disabled...\n")
  }

  # merge them back together
  new_line22 = paste(line22, collapse = "   ")

  # replace the old line with the new line
  file_cio[22] = new_line22

  # write the modified file.cio
  writeLines(text = file_cio, con = paste0(path, "file.cio"))

  # codes.sft modifications:
  # read in the codes.sft file (warnings turned off because of "incomplete final line")
  codes.sft = readLines(con = paste0(path, "codes.sft"), warn = F)

  # split that line based on white space
  line2 = codes.sft[2] %>% base::strsplit("\\s+") %>% unlist()

  # split line 3 based on white space
  line3 = codes.sft[3] %>% base::strsplit("\\s+") %>% unlist()

  # process for water balance routine:
  if (routine == "wb") {
    # find out what the column index is of HYD_HRU
    hyd_hru_col_index = which(line2 == "HYD_HRU")

    if (switch == "on") {
      # Edit ‘codes.sft’ file by changing the “n” to “y” in the hyd_hru column.
      line3[hyd_hru_col_index] = "y"
    } else{
      # Edit ‘codes.sft’ file by changing the “y” to “n” in the hyd_hru column.
      line3[hyd_hru_col_index] = "n"
    }
  }

  # process for crop yield routine
  if (routine == "cy") {
    # find out what the column index is of plnt
    plnt_index = which(line2 == "plnt")

    if (switch == "on") {
      # Edit ‘codes.sft’ file by changing the “n” to “y” in the plnt column.
      line3[plnt_index] = "y"
    } else{
      # Edit ‘codes.sft’ file by changing the “y” to “n” in the plnt column.
      line3[plnt_index] = "n"
    }
  }

  # merge line 3 back together and apply it to the file
  codes.sft[3] = paste(line3, collapse = "   ")

  # write the modified codes.sft file
  writeLines(text = codes.sft, con = paste0(path, "codes.sft"))
}

#' reads and re-formats the SWAT+ soft calibration output
#'
#' @param path string to (temporary) directory
#' @keywords internal
#' @importFrom data.table fread
#' @importFrom dplyr %>% mutate select
#' @importFrom purrr set_names
#' @importFrom readr read_lines
#' @importFrom stringr str_remove str_trim str_split
#' @importFrom tibble as_tibble
#' @importFrom tidyr unite
#'
#' @return returns a tibble of the formatted output of the soft-cal routine
#'
read_wb_aa <- function(path) {
  # read the wb_aa file from its path
  # fill = TRUE then in case the rows have unequal length, blank fields are
  # implicitly filled. (which is the case for this file)
  basin_wb_aa <- fread(paste0(path, "/basin_wb_aa.txt"), fill = TRUE, skip = 3)

  # change the column names to those of the second row in the text file.
  col_names <- read_lines(paste0(path, "/basin_wb_aa.txt"),
                          skip = 1, lazy = TRUE, n_max = 1) %>%
    str_trim(.) %>%
    str_split(., '[:space:]+') %>%
    unlist(.)
    # basin_wb_aa %>% slice(2) %>% unlist(., use.names = F)

  # some columns are not given a name (thanks..) need to figure out which ones
  # they are...
  no_name_columns <- seq(length(col_names) + 1, ncol(basin_wb_aa))

  basin_wb_aa <- basin_wb_aa %>%
    unite(., col = description, no_name_columns, sep = " ") %>%
    mutate(description = str_remove(description, ' NA')) %>%
    set_names(., c(col_names, 'description')) %>%
    as_tibble(.) %>%
    select(., precip, pet, et, eplant, esoil, surq_gen, latq, perc, qtile,
           wateryld, sw_init:sw_ave, sw_change, qtile, description)

  # return the formatted table
  return(basin_wb_aa)
}

#' Plot the water balance soft calibration iterations
#'
#' @param basin_wb_aa Table of avann water balance components results from
#'   softcal iterations
#'
#' @importFrom dplyr mutate select %>%
#' @importFrom purrr set_names
#' @importFrom readr read_lines
#' @importFrom stringr str_remove str_trim str_split
#' @importFrom tibble tibble
#' @importFrom tidyr pivot_longer
#' @import ggplot2
#' @import patchwork
#'
#' @keywords internal
#'
#' @return Returns a patchwork ggplot with softcal results
#'
plot_wb_softcal <- function(basin_wb_aa, wyr, bfr) {
  wyr_sim <- (basin_wb_aa$surq_gen +
                basin_wb_aa$latq +
                basin_wb_aa$perc +
                basin_wb_aa$qtile +
                basin_wb_aa$sw_change) / basin_wb_aa$precip

  bfr_sim <- (basin_wb_aa$perc + basin_wb_aa$latq) /
    (basin_wb_aa$surq_gen +
       basin_wb_aa$latq +
       basin_wb_aa$perc +
       basin_wb_aa$qtile)

  ratio_tbl <- tibble(iter = seq(length(wyr_sim),1,-1),
                      wyr  = wyr_sim,
                      bfr  = bfr_sim) %>%
    pivot_longer(., cols = - iter) %>%
    mutate(name = ifelse(name == 'wyr', 'Water yield', 'Base flow'))

  gg_ratio <- ggplot(data = ratio_tbl) +
    geom_line(aes(x = iter, y = value, col = name)) +
    geom_point(aes(x = iter, y = value, col = name, pch = name)) +
    geom_hline(yintercept = wyr, linetype = 2, color = 'steelblue4') +
    geom_hline(yintercept = bfr, linetype = 2, color = 'tomato4') +
    coord_flip() +
    scale_x_continuous(breaks = seq(length(wyr_sim),1,-1), labels = basin_wb_aa$description) +
    scale_color_manual(values = c('tomato4', 'steelblue4')) +
    labs(x = 'Iteration', y = 'Ratio', color = 'Ratio', pch = 'Ratio') +
    theme_bw() +
    theme(legend.position = 'bottom')

  wb_tbl <- basin_wb_aa %>%
    select(precip, pet, et, surq_gen, latq,perc, qtile) %>%
    mutate(iter = seq(nrow(.),1,-1)) %>%
    pivot_longer(., cols = -c(iter, precip, pet)) %>%
    mutate(name = factor(name, levels = c('et', 'surq_gen', 'latq', 'qtile', 'perc')))

  pcpet_tbl <- basin_wb_aa %>%
    select(precip, pet) %>%
    mutate(iter = seq(nrow(.),1,-1)) %>%
    pivot_longer(., cols = -c(iter)) %>%
    mutate(name = factor(name, levels = c('pet', 'precip')))

  gg_wb <- ggplot() +
    geom_col(data = wb_tbl, aes(x = iter, y = value, fill = name)) +
    geom_errorbarh(data = pcpet_tbl,
                   aes(xmax = iter + 0.5, xmin = iter - 0.5,
                       y = value, height = 0, col = name), linewidth = 1) +
    coord_flip() +
    scale_x_continuous(breaks = seq(length(wyr_sim),1,-1), labels = basin_wb_aa$description) +
    scale_fill_manual(values = c('skyblue2', 'steelblue4', 'springgreen3', 'wheat1', 'wheat4')) +
    scale_color_manual(values = c('tomato4', 'steelblue4')) +
    labs(x = '', y = 'Water balance (mm)', fill = 'Variable', color = ' ') +
    theme_bw() +
    theme(legend.position = 'bottom',
          axis.title.y = element_blank(),
          axis.text.y = element_blank())
  gg_ratio + gg_wb
}

#' Runs the SWAT+ soft-calibration routine for ....
#'
#' \code{run_softcal_cropyield} executes the SWAT+ soft calibration
#' routine to adjust the .... of a SWAT+ model setup.
#'
#' The user can specify the \code{crop_names} and the
#' \code{crop_yields} of the simulated watershed. Based on the provided
#' crops the routine adjusts the respective model parameters to improve
#' the fit of......
#'
#' @param project_path Path to the SWAT+ project folder (i.e. TxtInOut).
#' @param crop_yields Named vector, where the name is the crop name and the
#'   value defines the dry mass yield of the crop in t/ha.
#' @param start_date (optional) Start date of the SWAT simulation. Provided as
#'   character string in any ymd format (e.g. 'yyyy-mm-dd'), numeric value
#'   in the form yyyymmdd, or in Date format.
#' @param end_date (optional) End date of the SWAT simulation. Provided as
#'   character string in any ymd format (e.g. 'yyyy-mm-dd'), numeric value
#'   in the form yyyymmdd, or in Date format.
#' @param years_skip (optional) Integer value to define the number of simulation
#'   years that are skipped before writing SWAT model outputs.
#' @param keep_folder (optional) If \code{keep_folder = TRUE}
#'   '.model_run/verification' is kept and not deleted after finishing model runs.
#'   In this case '.model_run' is reused in a new model run if \code{refresh = FALSE}.
#'   \code{Default = FALSE}
#'
#' @return Returns the simulation results for the soft calibration routine as a
#'   tibble.
#' @importFrom processx run
#' @importFrom data.table fread
#' @importFrom tibble %>% tibble
#' @export
#'
run_softcal_cropyield <- function(project_path,
                                  crop_yields,
                                  start_date = NULL, end_date = NULL,
                                  years_skip = NULL,keep_folder = FALSE) {

  # Check settings before starting to set up
  ## General function input checks
  stopifnot(is.character(project_path))
  # stopifnot(is.character(crop_names))
  stopifnot(is.numeric(crop_yields))
  stopifnot(crop_yields >= 0)
  stopifnot(is.logical(keep_folder))

  # create a temporary directory copy of the model setup for crop yield
  print("creating temp model directory")
  run_path <- build_model_run(project_path, '/.run_softcal_cy')

  # enable the correct output
  set_print_prt(project_path, run_path, outputs = 'sft', years_skip)

  set_time_sim(project_path, run_path, start_date, end_date)

  # Build the water balance soft calibration input files and
  # write them into the path where the softcal will be run.
  build_cy_softcal_input(run_path, crop_yields)

  # enables the soft calibration routine
  toggle_sft(run_path, switch = "on", routine = "cy")

  ## Get operating system and find SWAT executable file
  os <- get_os()
  swat_exe <- find_swat_exe(project_path = run_path, os = os)

  print("running SWAT+ with soft-calibration routine")
  msg <- run(run_os(swat_exe, os),
             wd = run_path,
             error_on_status = FALSE)

  # reads the results of the cy soft calibration
  basin_crop_yld_aa = fread(paste0(run_path, "/basin_crop_yld_aa.txt"),
                            fill = TRUE,
                            skip = 1)

  # read the new .hyd file so that it can be returned
  hydrology_cal_hyd <- readLines(paste0(run_path, '/hydrology-cal.hyd'))

  #hydrology_cal_hyd <- hydrology_cal_hyd %>% setnames(c("NAME","LAT_TTIME","LAT_SED","CAN_MAX","ESCO","EPCO","ORGN_ENRICH","ORGP_ENRICH","CN3_SWF","BIO_MIX","PERCO","LAT_ORGN","LAT_ORGP","HARG_PET","LATQ_CO", "NULL"))


  # delete the temp directory if the user does not want to keep it.
  if (keep_folder == FALSE) {
    print("deleting temporary directory")
    unlink(run_path, recursive = TRUE, force = TRUE)
  }

  print("returning results..")

  return_data <- list(results = basin_crop_yld_aa %>% tibble(),
                      parameters = hydrology_cal_hyd)

  return(return_data)

  }

#' Build/load the crop yield soft calibration input files and write them
#' into the folder where the model is executed.
#'
#' @param run_path Path to the copy of SWAT+ project where the soft calibration
#'   will be executed
#' @param crop_yields Named vector, where the name is the crop name and the
#'   value defines the dry mass yield of the crop in t/ha.
#'   fraction of water yield which is base flow
#'
#' @importFrom readr read_lines write_lines
#' @importFrom purrr map map2 map2_df list_c list_rbind
#' @importFrom tibble tibble
#'
#' @keywords internal
#'
build_cy_softcal_input <- function(run_path, crop_yields) {

  # Remove existing soft calibration files in run folder
  invisible(suppressWarnings(file.remove(paste0(
    run_path,
    c('/plant_gro.sft',
      '/codes.sft',
      '/plant_parms.sft')
  ))))

  # Path where the SWATdoctR package was installed
  pkg_path <- system.file(package = "SWATdoctR")

  # Path to the water balance softcal input files
  cy_sft_path <- paste0(pkg_path, '/extdata/cy_softcal')

  # Build codes soft and write into the model run folder -----------------------
  codes_sft <- c('codes.sft: written by SWATdoctR',
                 'hyd_hru  hyd_hrulte  plnt  sed  nut  ch_sed  ch_nut  res',
                 '      n           n     y    n    n       n       n    n')

  write_lines(codes_sft, paste0(run_path, '/codes.sft'))

  # Build plant_gro.sft and write into the model run folder --------------------
  plant_gro_head <- c('plant_gro.sft: written by SWATdoctR',
                      '    1',
                      '  name         nlum  nspu',
                      paste0('  null         ',
                             sprintf('%4i', length(crop_yields)),
                             '    0'),
                     '  name             yield  npp  lai_mx  wstress  astress  tstress')

  plant_gro <- map2(names(crop_yields), crop_yields, ~ paste0('  ',
                                                              sprintf('%-16s', .x),
                                                              ' ',
                                                              sprintf('%5.2f', .y),
                                                              '    0       0        0        0        0')) %>%
    list_c(.)

  plant_gro <- c(plant_gro_head, plant_gro)

  write_lines(plant_gro, paste0(run_path, '/plant_gro.sft'))

  # Build plant_parms.sft and write into the model run folder ------------------
  plants_plt <- read_tbl(file = 'plants.plt', run_path = run_path, n_skip = 3)

  par_set_init  <- tibble(var = c('pest_stress', 'epco', 'lai_pot', 'harv_idx'),
                          name = NA_character_,
                          init = c(0.05, 1, NA, NA),
                          chg_typ = 'abschg',
                          neg = NA,
                          pos = c( 0.1,  1,  1, NA),
                          lo  = c(0.0099, 0, NA, NA),
                          up  = c(0.1, 1, NA,NA))

  plant_par_val <- map(names(crop_yields),
                       ~ parameterize_plant_parms(.x,
                                                  par_set_init,
                                                  plants_plt)) %>%
    list_rbind(.) %>%
    map2_df(., c('%-16s', '%-16s', '%16s', '%-8s', rep('%16s', 4)), ~ sprintf(.y, .x)) %>%
    apply(., 1, paste, collapse = '  ')

  plant_par_head <- c('plant_parms.sft: written by SWATdoctR',
                      '   1',
                      'name              plants  parms  nspu',
                      paste0('null              ',
                             sprintf('%6i', length(crop_yields)),
                             '      4     0'),
                      paste(
                        sprintf(c('%-16s', '%-16s', '%16s', '%-8s', rep('%16s', 4)),
                                colnames(par_set_init)),
                      collapse = '  '))

  plant_par <- c(plant_par_head, plant_par_val)
  write_lines(plant_par, paste0(run_path, '/plant_parms.sft'))
}


#' Parameterize the plant_parms.sft for a crop based on initial parameters in
#' plants.plt
#'
#' @param crop Crop name
#' @param par_set_init Initial parameter set
#' @param plants_plt Tibble with plant paramters from plants.plt
#'
#' @importFrom dplyr mutate %>%
#'
#' @keywords internal
parameterize_plant_parms <- function(crop, par_set_init, plants_plt) {
  par_set_init %>%
    mutate(name = crop,
           init = ifelse(var == 'lai_pot',
                         plants_plt$lai_pot[plants_plt$name == crop],
                         init),
           init = ifelse(var == 'harv_idx',
                         plants_plt$harv_idx[plants_plt$name == crop],
                         init),
           pos  = ifelse(var == 'harv_idx' & init >  1, 0.5, pos),
           pos  = ifelse(var == 'harv_idx' & init <= 1, 0.15, pos),
           neg  = - pos,
           lo   = ifelse(var %in% c('lai_pot' ,'harv_idx'), init + neg, init),
           lo   = ifelse(lo < 0, 0, lo),
           up   = ifelse(var %in% c('lai_pot' ,'harv_idx'), init + pos, init)
           )
}


#' Parameterize the plant_parms.sft for a crop based on initial parameters in
#' plants.plt
#'
#' @param project_path Path to the SWAT+ project folder (i.e. TxtInOut).
#' @param run_path Path to the copy of SWAT+ project where the soft calibration
#'   will be executed
#'
#' @importFrom dplyr %>% nth
#' @importFrom stringr str_replace
#'
#' @export
accept_softcal_changes <- function(project_path, hyd_file){

  # write the passed hyd_file to the project directory.
  writeLines(text = hyd_file, con = paste0(project_path, "/hydrology-softcal.hyd"))

  # read in the file.cio from the project path
  file_cio <- readLines(con = paste0(project_path, "/file.cio"))

  # the modifications made to file.cio will depend on whether or not the soft
  # calibration changes have already been changed (IF we want to keep the soft
  # cal results along with the original...)

  hyd_file_name <- file_cio[17] %>% base::strsplit("\\s+") %>% unlist() %>% dplyr::nth(2)

  # if the file has already been modified by SWATdoctR:
  if (hyd_file_name == "hydrology-softcal.hyd") {

    # you wont have to update file.cio because it  will have already been changed...
    print("please note! file.cio in your project directory was modified by SWATdoctR previously, so it was not altered again")
    print("hydrology-softcal.hyd is already in use -- and has now been updated with the latest parameters")

  } else if (hyd_file_name == "hydrology.hyd") {

    # if the file.cio has not been modified yet, then we can make our modifications normally.
    file_cio[1] <- paste(file_cio[1], "and modified by SWATdoctR [[version?]]] on", Sys.time())

    # I hope this works, i have no idea how this file.cio works (like how are the different columns used?)
    file_cio[17] <- file_cio[17] %>% stringr::str_replace("hydrology.hyd", "hydrology-softcal.hyd")

    # upate file.cio with changes
    writeLines(text = file_cio, con = paste0(project_path, "/file.cio"))

    print("original SWAT project now using hydrology parameters from soft calibration.")
    print("please note! file.cio in your project directory has been adjusted to use hydrology-cal.hyd (line 17, column 2)")

  } else{
    # not sure how to handle this case (neither hydrology.hyd nor hydrology-cal.hyd present in line 17, col 2)
    # doing a warning for now...
    warning("file name in line 17, column 2 not recognized. please change to hydrology.hyd or hydrology.hyd")
  }
}
